use cgmath::{Deg, Matrix4, Point2, Vector3};

use super::OPENGL_TO_WGPU_MATRIX;

// Represents a sprite tranformation.
//
// This produces a single matrix we can apply to a brand-new sprite instance
// to put it in the right place, at the right angle, with the right scale.
pub struct Transform {
	pub pos: Point2<f32>,   // position on screen
	pub screen_aspect: f32, // width / height. Screen aspect ratio.
	pub aspect: f32,        // width / height. Sprite aspect ratio.
	pub scale: f32,         // if scale = 1, this sprite will be as tall as the screen.
	pub rotate: Deg<f32>,   // Around this object's center, in degrees measured ccw from vertical
}

impl Transform {
	/// Build the matrix that corresponds to this transformation.
	pub fn to_matrix(&self) -> Matrix4<f32> {
		// Note that our mesh starts centered at (0, 0).
		// This is essential---we do not want scale and rotation
		// changing our sprite's position!

		// Apply sprite aspect ratio, preserving height.
		// This must be done *before* rotation.
		//
		// We apply the provided scale here as well as a minor optimization
		let sprite_aspect_and_scale =
			Matrix4::from_nonuniform_scale(self.aspect * self.scale, self.scale, 1.0);

		// Apply rotation
		let rotate = Matrix4::from_angle_z(self.rotate);

		// Apply screen aspect ratio, again preserving height.
		// This must be done AFTER rotation... think about it!
		let screen_aspect = Matrix4::from_nonuniform_scale(1.0 / self.screen_aspect, 1.0, 1.0);

		// After finishing all op, translate.
		// This must be done last, all other operations
		// require us to be at (0, 0).
		let translate = Matrix4::from_translation(Vector3 {
			x: self.pos.x,
			y: self.pos.y,
			z: 0.0,
		});

		// Order matters!
		// The rightmost matrix is applied first.
		return OPENGL_TO_WGPU_MATRIX
			* translate * screen_aspect
			* rotate * sprite_aspect_and_scale;
	}
}
